<?php

declare(strict_types=1);

namespace mirzaev\vk\api\methods;

// Файлы проекта
use mirzaev\vk\robots\robot,
    mirzaev\vk\robots\group;

// Встроенные библиотеки
use stdClass,
    Exception;

/**
 * Режимы отправки сообщений
 */
enum message_send
{
    /** Обычная отправка */
    case simple;

    /** Проверка отправки */
    case check;
}

/**
 * Режимы генерации идентификатора сессии доставки сообщения
 */
enum random_id
{
    /** Генерация: time() (нельзя отправить более чем 1 сообщение в секунду) */
    case date;

    /** Генерация: rand() */
    case random;

    /** Генерация: random_int(0, PHP_INT_MAX) */
    case crypto;
}

/**
 * Сообщение
 *
 * @method public static function put(RobotAbstract $from, int $to, string $message, int $mode = 2) Отправить сообщение
 *
 * @see https://vk.com/dev/messages.send
 * @see https://vk.com/dev/messages.getById
 *
 * @package mirzaev\vk\api
 * @author Arsen Mirzaev Tatyano-Muradovich <arsen@mirzaev.sexy>
 *
 * @todo Доработать строгий режим отправки: проверку сообщения в беседе (не имеет ID сообщений)
 */
final class messages extends method
{
    /**
     * @var message_send $send_mode Режим отправки сообщений
     */
    protected message_send $send_mode = message_send::simple;

    /**
     * @var random_id $generate_mode Режим генерации идентификатора сессии доставки сообщения
     */
    protected random_id $generate_mode = random_id::crypto;

    /**
     * @var int|string|array|null $peer_id Получатель
     */
    protected int|string|array|null $peer_id = null;

    /**
     * @var string|null $text Текст
     */
    protected string|null $text = null;

    /**
     * @var ?int $lat Географическая ширина
     */
    protected ?int $lat = null;

    /**
     * @var ?int $long Географическая долгота
     */
    protected ?int $long = null;

    /**
     * @var ?int $reply_to Идентификатор сообщения, на которое требуется ответить
     */
    protected ?int $reply_to = null;

    /**
     * @var ?array $forward_messages Идентификаторы пересылаемых сообщений
     */
    protected ?array $forward_messages = null;

    /**
     * @var ?string $sticker_id Идентификатор стикера
     */
    protected ?string $sticker_id = null;

    /**
     * @var ?string $payload Полезная нагрузка
     */
    protected ?string $payload = null;

    /**
     * @var bool $dont_parse_links Не создавать представление ссылки в сообщении?
     */
    protected bool $dont_parse_links = false;

    /**
     * @var bool $disable_mentions Отключить уведомление об упоминании в сообщении?
     */
    protected bool $disable_mentions = false;

    /**
     * @var ?string $intent Интент
     */
    protected ?string $intent = null;

    /**
     * @var ?int $subscribe_id Число, которое будет использоваться для работы с интентами
     */
    protected ?int $subscribe_id = null;

    /**
     * @var ?int $offset Смещение для выбора сообщений
     */
    protected ?int $offset = null;

    /**
     * @var ?int $count Количество выбираемых сообщений
     */
    protected ?int $count = null;

    /**
     * @var ?int $start_message_id Идентификатор сообщения от которого начнётся выборка
     */
    protected ?int $start_message_id = null;

    /**
     * @var bool $rev Получить сообщения в хронологичном порядке? (иначе в обратном)
     */
    protected bool $rev = false;

    /**
     * @var bool $extended Получить расширенную информацию об авторах сообщений?
     */
    protected bool $extended = false;

    /**
     * @var ?array $fields Дополнительные запрашиваемые поля (для $extended)
     */
    protected ?array $fields = null;

    /**
     * Конструктор
     *
     * @param robot $robot Робот
     */
    public function __construct(protected robot $robot, mixed ...$properties)
    {
        // Быстрая отправка
        if (!empty($properties))  $this->send(...$properties);
    }

    /**
     * Записать: режим отправки сообщений
     *
     * @param message_send $send_mode Режим отправки сообщений
     *
     * @return static
     */
    public function send_mode(message_send $send_mode): static
    {
        // Запись в свойство
        $this->__set('send_mode', $send_mode);

        return $this;
    }

    /**
     * Записать: режим генерации идентификатора сессии доставки сообщения
     *
     * @param random_id $generate_mode Режим генерации идентификатора сессии доставки сообщения
     *
     * @return static
     */
    public function generate_mode(random_id $generate_mode): static
    {
        // Запись в свойство
        $this->__set('generate_mode', $generate_mode);

        return $this;
    }

    /**
     * Записать: получатель
     *
     * @param int|string|array|null $peer_id Получатель
     *
     * @return static
     */
    public function peer_id(int|string|array|null $peer_id): static
    {
        // Запись в свойство
        $this->__set('peer_id', $peer_id);

        return $this;
    }

    /**
     * Записать: текст
     *
     * @param string|null $text Текст
     *
     * @return static
     */
    public function text(string|null $text): static
    {
        // Запись в свойство
        $this->__set('text', $text);

        return $this;
    }

    /**
     * Записать: географическая ширина
     *
     * @param ?int $lat Географическая ширина
     *
     * @return static
     */
    public function lat(?int $lat): static
    {
        // Запись в свойство
        $this->__set('lat', $lat);

        return $this;
    }

    /**
     * Записать: географическая долгота
     *
     * @param ?int $long Географическая долгота
     *
     * @return static
     */
    public function long(?int $long): static
    {
        // Запись в свойство
        $this->__set('long', $long);

        return $this;
    }

    /**
     * Записать: идентификатор сообщения, на которое требуется ответить
     *
     * @param ?int $reply_to Идентификатор сообщения, на которое требуется ответить
     *
     * @return static
     */
    public function reply_to(?int $reply_to): static
    {
        // Запись в свойство
        $this->__set('reply_to', $reply_to);

        return $this;
    }

    /**
     * Записать: идентификаторы пересылаемых сообщений
     *
     * @param ?array $forward_messages Идентификаторы пересылаемых сообщений
     *
     * @return static
     */
    public function forward_messages(?array $forward_messages): static
    {
        // Запись в свойство
        $this->__set('forward_messages', $forward_messages);

        return $this;
    }

    /**
     * Записать: идентификаторы пересылаемых сообщений
     *
     * @param ?array $forward Идентификаторы пересылаемых сообщений
     *
     * @return static
     */
    public function forward(?array $forward): static
    {
        // Запись в свойство
        $this->__set('forward_messages', $forward);

        return $this;
    }

    /**
     * Записать: идентификатор стикера
     *
     * @param ?string $sticker_id Идентификатор стикера
     *
     * @return static
     */
    public function sticker_id(?string $sticker_id): static
    {
        // Запись в свойство
        $this->__set('sticker_id', $sticker_id);

        return $this;
    }

    /**
     * Записать: полезная нагрузка
     *
     * @param ?string $payload Полезная нагрузка
     *
     * @return static
     */
    public function payload(?string $payload): static
    {
        // Запись в свойство
        $this->__set('payload', $payload);

        return $this;
    }

    /**
     * Записать: не создавать представление ссылки в сообщении?
     *
     * @param bool $dont_parse_links Не создавать представление ссылки в сообщении?
     *
     * @return static
     */
    public function dont_parse_links(bool $dont_parse_links): static
    {
        // Запись в свойство
        $this->__set('dont_parse_links', $dont_parse_links);

        return $this;
    }

    /**
     * Записать: отключить уведомление об упоминании в сообщении?
     *
     * @param bool $disable_mentions Отключить уведомление об упоминании в сообщении?
     *
     * @return static
     */
    public function disable_mentions(bool $disable_mentions): static
    {
        // Запись в свойство
        $this->__set('disable_mentions', $disable_mentions);

        return $this;
    }

    /**
     * Записать: интент
     *
     * @param ?string $intent Интент
     *
     * @return static
     */
    public function intent(?string $intent): static
    {
        // Запись в свойство
        $this->__set('intent', $intent);

        return $this;
    }

    /**
     * Записать: число, которое будет использоваться для работы с интентами
     *
     * @param ?int $subscribe_id Число, которое будет использоваться для работы с интентами
     *
     * @return static
     */
    public function subscribe_id(?int $subscribe_id): static
    {
        // Запись в свойство
        $this->__set('subscribe_id', $subscribe_id);

        return $this;
    }

    /**
     * Записать: смещение для выбора сообщений
     *
     * @param ?int $offset Смещение для выбора сообщений
     *
     * @return static
     */
    public function offset(?int $offset): static
    {
        // Запись в свойство
        $this->__set('offset', $offset);

        return $this;
    }

    /**
     * Записать: количество выбираемых сообщений
     *
     * @param ?int $count Количество выбираемых сообщений
     *
     * @return static
     */
    public function count(?int $count): static
    {
        // Запись в свойство
        $this->__set('count', $count);

        return $this;
    }

    /**
     * Записать: идентификатор сообщения от которого начнётся выборка
     *
     * @param ?int $start_message_id Идентификатор сообщения от которого начнётся выборка
     *
     * @return static
     */
    public function start_message_id(?int $start_message_id): static
    {
        // Запись в свойство
        $this->__set('start_message_id', $start_message_id);

        return $this;
    }

    /**
     * Записать: получить сообщения в хронологичном порядке?
     *
     * @param bool $rev Получить сообщения в хронологичном порядке? (иначе в обратном)
     *
     * @return static
     */
    public function rev(bool $rev): static
    {
        // Запись в свойство
        $this->__set('rev', $rev);

        return $this;
    }

    /**
     * Записать: получить расширенную информацию об авторах сообщений?
     *
     * @param bool $extended Получить расширенную информацию об авторах сообщений?
     *
     * @return static
     */
    public function extended(bool $extended): static
    {
        // Запись в свойство
        $this->__set('extended', $extended);

        return $this;
    }

    /**
     * Записать: дополнительные запрашиваемые поля
     *
     * @param ?array $fields Дополнительные запрашиваемые поля (для static::$extended)
     *
     * @return static
     */
    public function fields(?array $fields): static
    {
        // Запись в свойство
        $this->__set('fields', $fields);

        return $this;
    }

    /**
     * Инициализировать получателя
     *
     * Записывает получателя в буфер API робота
     *
     * @param int|string|array|null $peer_id Получатель (группа, чат, пользователь и пользователи)
     *
     * @return void
     */
    public function receiver(int|string|array|null $peer_id = null): void
    {
        if (isset($peer_id)) $this->peer_id = $peer_id;
        if (is_int($this->peer_id)) ($id = $this->peer_id - 2000000000) > 0 ? $this->robot->api['peer_id'] = $this->peer_id : $this->robot->api['chat_id'] = $id;
        else if (is_array($this->peer_id)) $this->robot->api['peer_ids'] = implode(',', $this->peer_id);
        else if (is_string($this->peer_id)) $this->robot->api['domain'] = $this->peer_id;
    }

    /**
     * Отправить сообщение
     *
     * @see https://dev.vk.com/method/messages.send
     *
     * @param int|string|null $peer_id Получатель (группа, чат, пользователь и пользователи)
     * @param ?string $message Сообщение
     * @param ?int $lat Географическая ширина
     * @param ?int $long Географическая долгота
     * @param ?array $attachments Вложения
     * @param ?int $reply_to Идентификатор сообщения, на которое требуется ответить
     * @param ?array $forward_messages Идентификаторы пересылаемых сообщений
     * @param ?forward $forward Пересылаемые сообщения (в другой чат)
     * @param ?string $sticker_id Идентификатор стикера
     * @param ?keyboard $keyboard Инстанция клавиатуры
     * @param ?template $template Инстанция шаблона сообщения
     * @param ?string $payload Полезная нагрузка
     * @param bool $dont_parse_links Не создавать представление ссылки в сообщении?
     * @param bool $disable_mentions Отключить уведомление об упоминании в сообщении?
     * @param ?string $intent Интент
     * @param ?int $subscribe_id Число, которое будет использоваться для работы с интентами
     * @param ?int $random_id Идентификатор сессии доставки сообщения (защита от повторных отправок)
     *
     * @return int|array Идентификатор успешно отправленного сообщения или ответ сервера (подразумевается ошибка)
     *
     * @todo
     * 1. Написать обработчик ошибок возвращаемых ВКонтакте
     * 2. Добавить параметр forward (не путать с forward_messages)
     * 3. Добавить клавиатуру
     * 4. Добавить шаблоны сообщений
     * 5. Добавить content_source
     */
    public function send(
        int|string|null $peer_id = null,
        ?string $message = null,
        ?int $lat = null,
        ?int $long = null,
        ?array $attachments = null,
        ?int $reply_to = null,
        ?array $forward_messages = null,
        // ?forward $forward = null,
        ?string $sticker_id = null,
        // ?keyboard $keyboard = null,
        // ?template $template = null,
        ?string $payload = null,
        bool $dont_parse_links = false,
        bool $disable_mentions = false,
        ?string $intent = null,
        ?int $subscribe_id = null,
        ?int $random_id = null,
    ): int|array {
        // Реинициализация настроек
        $this->robot->api->reinit();

        // Инициализация получателя
        $this->receiver($peer_id);

        // Инициализация идентификатора сессии доставки сообщения (защита от повторных отправок)
        $this->robot->api['random_id'] = $random_id ?? match ($this->generate_mode) {
            random_id::date => time(),
            random_id::random => rand(),
            random_id::crypto => random_int(0, PHP_INT_MAX),
            default => random_int(0, PHP_INT_MAX)
        };

        // Инициализация текста в настройках API
        if (!empty($message ??= $this->text)) $this->robot->api['message'] = $message;

        // Инициализация широты
        if (!empty($lat ??= $this->lat)) $this->robot->api['lat'] = $lat;

        // Инициализация долготы
        if (!empty($long ??= $this->long)) $this->robot->api['long'] = $long;

        // Инициализация вложений
        if (!empty($attachments ??= $this->robot->api->data)) $this->robot->api['attachment'] = implode(',', $attachments);

        // Инициализация сообщения, на которое требуется ответить
        if (!empty($reply_to ??= $this->reply_to)) $this->robot->api['reply_to'] = $reply_to;

        // Инициализация пересылаемых сообщений
        if (!empty($forward_messages ??= $this->forward_messages)) $this->robot->api['forward_messages'] = implode(',', $forward_messages);

        // Инициализация стикера
        if (!empty($sticker_id ??= $this->sticker_id)) $this->robot->api['sticker_id'] = $sticker_id;

        // Инициализация полезной нагрузки
        if (!empty($payload ??= $this->payload)) $this->robot->api['payload'] = $payload;

        // Инициализация пользовательского соглашения
        // $this->robot->api['content_source'] = $this->robot->content_source;

        // Инициализация "не создавать представление ссылки в сообщении?"
        if ($dont_parse_links || $this->dont_parse_links) $this->robot->api['dont_parse_links'] = 1;

        // Инициализация "отключить уведомление об упоминании в сообщении?"
        if ($disable_mentions || $this->disable_mentions) $this->robot->api['disable_mentions'] = 1;

        // Инициализация интентов
        if (!empty($intent ??= $this->intent)) $this->robot->api['intent'] = $intent;

        // Инициализация числа, которое будет использоваться для работы с интентами
        if (!empty($subscribe_id ??= $this->subscribe_id)) $this->robot->api['subscribe_id'] = $subscribe_id;

        // Проверка сформированного сообщения
        if (!$this->robot->api->offsetExists('message') && !$this->robot->api->offsetExists('attachment')) throw new Exception('Сообщение должно содержать текст, либо вложение');

        // Запрос
        $request = json_decode($this->robot->browser->request('POST', 'messages.send', ['form_params' => $this->robot->api->settings])->getBody()->getContents());

        // Проверка на наличие ошибок в ответе от ВКонтакте
        if (isset($request->error)) throw new Exception('ВКонтакте: ' . $request->error->error_msg, $request->error->error_code);

        if ($this->send_mode === message_send::check) {
            // Надёжная доставка сообщения

            if (!empty($request["response"])) {
                // Ответ получен

                // Реиницилазиция
                $this->robot->api->reinit();

                // Запрашиваемые сообщения
                $this->robot->api['message_ids'] = $request["response"];

                if ($this->robot->browser->request('POST', 'messages.getById', ['form_params' => $this->robot->api->settings])['response']['count'] === 0) {
                    // Сообщения не существует
                    //!!!!!!!!!!!!!!!!!!!!!!!!!!!
                    //!!!!!!!!!!!!!!!!!!!!!!!!!!
                    //!!!!!!!!!!!!!!!!!!!!!!!!!

                    // Повторная отправка
                    $this->send();
                }
            } else {
            }
        }

        return $request->response;
    }

    /**
     * Удалить сообщение
     *
     * @see https://dev.vk.com/method/messages.delete
     *
     * @param int|string|array|null $messages Получатель (message_ids + cmids)
     * @param int|string|null $peer_id Идентификатор беседы
     * @param bool $spam Пометить как спам?
     * @param bool $delete_for_all Удалить для всех?
     *
     * @return int|array Идентификатор успешно отправленного сообщения или ответ сервера (подразумевается ошибка)
     *
     * @todo
     * 1. Переделать под новую версию
     * 2. Написать обработчик ошибок возвращаемых ВКонтакте
     */
    public function delete(string|int|array|null $messages = null, int|string|null $peer_id = null, bool $spam = false, bool $delete_for_all = false): stdClass
    {
        // Реиницилазиция настроек
        $this->robot->api->reinit();

        if (isset($peer_id)) {
            // Получен идентификатор беседы

            // Инициализация идентификатора беседы
            $this->robot->api['peer_id'] = $peer_id;

            // Инициализация идентификаторов сообщений
            $this->robot->api['cmids'] = $messages;

            // Инициализация: "удалить для всех?"
            $this->robot->api['delete_for_all'] = 1;
        } else {
            // Не получен идентификатор беседы

            // Инициализация идентификаторов сообщений
            $this->robot->api['message_ids'] = $messages;

            // Инициализация: "удалить для всех?"
            if ($delete_for_all) $this->robot->api['delete_for_all'] = 1;
        }

        // Инициализация: "сообщить о спаме?"
        if ($spam) $this->robot->api['spam'] = $spam;

        // Запрос
        $request = json_decode($this->robot->browser->request('POST', 'messages.delete', ['form_params' => $this->robot->api->settings])->getBody()->getContents());

        // Проверка на наличие ошибок в ответе от ВКонтакте
        if (isset($request->error)) throw new Exception('ВКонтакте: ' . $request->error->error_msg, $request->error->error_code);


        return $request->response;
    }

    /**
     * Получить информацию о сообщении по id
     *
     * @param string $message_ids Идентификатор сообщения
     *
     * @return object Информация о сообщении
     *
     * @todo
     * 1. Доделать, а лучше полностью переделать
     */
    public function getById(int $message_ids): object
    {
        // Реиницилазиция
        $this->robot->api->reinit();

        // Идентификаторы сообщений. Максимум 100 идентификаторов
        $this->robot->api['message_ids'] = $message_ids;


        if ($this->robot instanceof group) {
            // Если это группа

            // Идентификатор группы
            isset($this->robot->id) ? $this->robot->api['group_id'] = $this->robot->id : throw new Exception('Необходимо указать идентификатор сообщества');
        }

        // Запрос
        $request = json_decode(($this->robot->browser->request('POST', 'messages.getById', ['form_params' => $this->robot->api->settings])->getBody()->getContents()));

        // Проверка на наличие ошибок в ответе от ВКонтакте
        if (isset($request->error)) throw new Exception('ВКонтакте: ' . $request->error->error_msg, $request->error->error_code);

        return $request->response;
    }

    /**
     * Получить историю сообщений
     *
     * @see https://dev.vk.com/method/messages.getHistory
     *
     * @param ?int $offset Смещение для выбора сообщений
     * @param ?int $count Количество выбираемых сообщений
     * @param int|string|null $peer_id Получатель (группа, чат, пользователь и пользователи)
     * @param ?int $start_message_id Идентификатор сообщения от которого начнётся выборка
     * @param bool $rev Получить сообщения в хронологичном порядке? (иначе в обратном)
     * @param bool $extended Получить расширенную информацию об авторах сообщений?
     * @param ?array $fields Дополнительные запрашиваемые поля (для $extended)
     *
     * @return int|array Идентификатор успешно отправленного сообщения или ответ сервера (подразумевается ошибка)
     *
     * @todo Написать обработчик ошибок возвращаемых ВКонтакте
     */
    public function getHistory(
        ?int $offset = null,
        ?int $count = null,
        int|string|null $peer_id = null,
        ?int $start_message_id = null,
        bool $rev = false,
        bool $extended = false,
        ?array $fields = null
    ): object {
        // Реиницилазиция
        $this->robot->api->reinit();

        // Инициализация смещения для выбора сообщений
        if (!empty($offset ??= $this->offset)) $this->robot->api['offset'] = $offset;

        // Инициализация количества выбираемых сообщений
        if (!empty($count ??= $this->count)) $this->robot->api['count'] = $count;

        // Инициализация получателя
        $this->receiver($peer_id);

        // Инициализация идентификатора сообщения от которого начнётся выборка
        if (!empty($start_message_id ??= $this->start_message_id)) $this->robot->api['start_message_id'] = $start_message_id;

        // Инициализация "получить сообщения в хронологичном порядке?"
        if ($rev || $this->rev) $this->robot->api['rev'] = 1;

        // Инициализация "получить расширенную информацию об авторах сообщений?"
        if ($extended || $this->extended) $this->robot->api['extended'] = 1;

        if (!empty($fields ??= $this->fields)) {
            // Инициализация дополнительных запрашиваемых полей

            // Инициализация значения в буфере запроса к API
            $this->robot->api['fields'] = '';

            foreach ($fields as $key => $field) {
                // Перебор дополнительных запрашиваемых полей

                // Запись в буфер запроса к API
                $this->robot->api['fields'] .= $field;

                // Запись разделителя
                if ($key !== array_key_last($fields)) $this->robot->api['fields'] .= ', ';
            }
        }

        // Запрос
        $request = json_decode(($this->robot->browser->request('POST', 'messages.getHistory', ['form_params' => $this->robot->api->settings])->getBody()->getContents()));

        // Проверка на наличие ошибок в ответе от ВКонтакте
        if (isset($request->error)) throw new Exception('ВКонтакте: ' . $request->error->error_msg, $request->error->error_code);

        return $request->response;
    }

    /**
     * Записать свойство
     *
     * @param string $name Название
     * @param mixed $value Значение
     *
     * @return void
     */
    public function __set(string $name, mixed $value): void
    {
        match ($name) {
            'send_mode' => $this->send_mode = $value,
            'generate_mode' => $this->generate_mode = $value,
            'peer_id', 'destination' => $this->peer_id = $value,
            'text' => $this->text = $value,
            'lat', 'latitude' => $this->lat = $value,
            'long', 'longitude' => $this->long = $value,
            'reply_to', 'reply', 'to' => $this->reply_to = $value,
            'forward_messages' => $this->forward_messages = $value,
            'sticker_id', 'sticker' => $this->sticker_id = $value,
            'payload' => $this->payload = $value,
            'dont_parse_links' => $this->dont_parse_links = $value,
            'parse_links', 'parse' => $this->dont_parse_links = !$value,
            'disable_mentions' => $this->disable_mentions = $value,
            'mentions' => $this->disable_mentions = !$value,
            'intent' => $this->intent = $value,
            'subscribe_id' => $this->subscribe_id = $value,
            'offset' => $this->offset = $value,
            'count', 'amount' => $this->count = $value,
            'start_message_id' => $this->start_message_id = $value,
            'rev' => $this->rev = $value,
            'extended' => $this->extended = $value,
            'fields' => $this->fields = $value,
            default => throw new Exception("Свойство $name не найдено", 404)
        };
    }

    /**
     * Прочитать свойство
     *
     * @param string $name Название
     *
     * @return mixed
     */
    public function __get(string $name): mixed
    {
        return match ($name) {
            'send_mode' => $this->send_mode,
            'generate_mode' => $this->generate_mode,
            'peer_id', 'destination' => $this->peer_id,
            'text' => $this->text,
            'lat', 'latitude' => $this->lat,
            'long', 'longitude' => $this->long,
            'reply_to', 'reply', 'to' => $this->reply_to,
            'forward_messages' => $this->forward_messages,
            'sticker_id', 'sticker' => $this->sticker_id,
            'payload' => $this->payload,
            'dont_parse_links', 'parse_links', 'parse' => $this->dont_parse_links,
            'disable_mentions', 'mentions' => $this->disable_mentions,
            'intent' => $this->intent,
            'subscribe_id' => $this->subscribe_id,
            'offset' => $this->offset,
            'count', 'amount' => $this->count,
            'start_message_id' => $this->start_message_id,
            'rev' => $this->rev,
            'extended' => $this->extended,
            'fields' => $this->fields,
            default => throw new Exception("Свойство $name не найдено", 404)
        };
    }
}
